/* pushme.c V1.0 QRT090298 */
/* email: kurt.schuster@t-online.de */
/* home: http://home.t-online.de/home/kurt.schuster/ */

#pragma pack(2)!
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include "pushme.h"

#define PushMeAppID         'PuMe'   
#define PushMeDBType        'Data'  

static int StartApplication(void);
static Boolean OpenDatabase(void);
static void EventLoop(void);
static void StopApplication(void);
static Boolean FormEventGame(EventPtr event);

DmOpenRef PushMeDB;
char PushMeDBName[] = "PushMeDB";

//  char debugstring1[20];
//	StrIToA(debugstring1, MyNumber)
//	ErrFatalDisplayIf( true, debugstring1);

#define MAXS    10
#define PFTOP    3
#define PFMAXW 155
#define PFMAXH 155

static void GenPlayField(void);
static void DrawTile(int, int, int);
static void TilesAskSize(void);
static void TilesMix(void);
static void DetTile(int, int);
static void DetMoveTiles(int, int);
static void MoveTile(int, int);
static void DrawTileFrame(int, int);
static int sign(int);

#define RECSIZE 2*sizeof(int)+MAXS*MAXS*sizeof(Byte)

struct tile{
    RectangleType r;
    Byte digit;
};

struct tile t[MAXS][MAXS];
UInt index = 0;
Boolean newpf = true;
int tilesx = 4, tilesy = 4;   // playfieldsize in tiles
int holex, holey;             // holeposition
int twidth, thight, pfwidth, pfhight, pfleft;   // in pixels

//*****************************************************************
//
// PilotMain as usual
//
//*****************************************************************

DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
int error;

    if(cmd == sysAppLaunchCmdNormalLaunch){
        error = StartApplication();
        if(error) return error;
        EventLoop();
        StopApplication();
    }
    return 0;
}

//*****************************************************************
//
// StartApplication as usual
//
//*****************************************************************

static int StartApplication(void)
{
int error;

    error = OpenDatabase();
    if(error) return error;
    FrmGotoForm(formID_game);
}

//*****************************************************************
// OpenDatabase
// create one if none exists, else open it and put it's data into
// the tile struct
//*****************************************************************

static Boolean OpenDatabase(void)
{
int x, y;
int  *ip;
Byte *bp;
VoidHand RecHandle;
VoidPtr RecPtr;

    PushMeDB = DmOpenDatabaseByTypeCreator(PushMeDBType, PushMeAppID, dmModeReadWrite);
    if(!PushMeDB){
        if(DmCreateDatabase(0, PushMeDBName, PushMeAppID, PushMeDBType, false)) return 1;
        PushMeDB = DmOpenDatabaseByTypeCreator(PushMeDBType, PushMeAppID, dmModeReadWrite);
        RecHandle = DmNewRecord(PushMeDB, &index, RECSIZE);
    }
    else
    {
        newpf = false;
        
        RecHandle = DmGetRecord(PushMeDB, index);
        RecPtr =  MemHandleLock(RecHandle);
        
        ip = (int *)RecPtr;
        bp = (Byte *)(RecPtr + 2*sizeof(int));

        tilesx = *ip++; 
        tilesy = *ip;
        
        for(x=0; x<MAXS; x++){
            for(y=0; y<MAXS; y++){
                t[x][y].digit = *bp++;
            }
        }

        MemPtrUnlock(RecPtr);        
    }
    
    DmReleaseRecord(PushMeDB, index, true);

    return 0;
}

//*****************************************************************
//
// EventLoop as usual
//
//*****************************************************************

static void EventLoop(void)
{
short err;
int formID;
FormPtr form;
EventType event;

    do{
        EvtGetEvent(&event, evtWaitForever);

        if(SysHandleEvent(&event)) continue;
        if(MenuHandleEvent((void *)0, &event, &err)) continue;

        if(event.eType == frmLoadEvent){
            formID = event.data.frmLoad.formID;
            form = FrmInitForm(formID);
            FrmSetActiveForm(form);
            
            switch(formID){
                case formID_game:
                    FrmSetEventHandler(form, (FormEventHandlerPtr) FormEventGame);
                    break;
            }
        }
        FrmDispatchEvent(&event); 
    }while(event.eType != appStopEvent);   
}

//*****************************************************************
// StopApplication
// write parts of tile struct to database and close it
//*****************************************************************

static void StopApplication(void)
{
int x, y;
ULong o;
VoidHand RecHandle;
VoidPtr RecPtr;

    RecHandle = DmGetRecord(PushMeDB, index);
    RecPtr =  MemHandleLock(RecHandle);
    
    DmWrite(RecPtr, 0, &tilesx, 2);
    DmWrite(RecPtr, 2, &tilesy, 2);
    
    o=4;
    for(x=0; x<MAXS; x++){
        for(y=0; y<MAXS; y++){
            DmSet(RecPtr, o++, 1, t[x][y].digit);
        }
    }

    MemPtrUnlock(RecPtr);        
    DmReleaseRecord(PushMeDB, index, true);  
    DmCloseDatabase(PushMeDB);
}

//*****************************************************************
//
// FormEventGame as usual
//
//*****************************************************************

static Boolean FormEventGame(EventPtr event)
{
FormPtr form;
int handled = false;

    switch(event->eType){
        case frmOpenEvent:
            form = FrmGetActiveForm();
            FrmDrawForm(form);
            GenPlayField();
            handled = true;
            break;
        case penDownEvent:
            DetTile(event->screenX, event->screenY);
            handled = true;
            break;
        case ctlSelectEvent:  // A control button was pressed and released.
            switch(event->data.ctlEnter.controlID){
                // case buttonID_editdone:
                //    handled = true;
                //    break;
            }
            break;
        case menuEvent:
            MenuEraseStatus(NULL);
            switch (event->data.menu.itemID){
                case menuitemID_size:
                    TilesAskSize();
                    handled = true;
                    break;
                case menuitemID_mix:
                    TilesMix();
                    handled = true;
                    break;
                case menuitemID_about:
                    FrmAlert(alertID_about);
                    handled = true;
                    break;
            }
            break;
    }
    return handled;
}

//*****************************************************************
// GenPlayField
// generates the playfield, fills the tile struct, datas from  
// a database are regarded
//*****************************************************************

static void GenPlayField()
{
int x, y, d;
RectangleType r;

    twidth = PFMAXW / tilesx;
    thight = PFMAXH / tilesy;
    pfwidth = twidth * tilesx;
    pfhight = thight * tilesy;    
    pfleft = (160-pfwidth) / 2;

    r.topLeft.x = pfleft - 1;
    r.topLeft.y = PFTOP - 1;
    r.extent.x = pfwidth + 1;
    r.extent.y = pfhight + 1;
    
    WinEraseWindow();
    // FrmUpdateForm(formID_game, frmRedrawUpdateCode); // restores the menubar
    WinDrawRectangleFrame(boldRoundFrame, &r);

    if(thight>20 && twidth>20)
        FntSetFont(largeFont);
    else
        FntSetFont(stdFont);

    d = 1;
    
    for(y=0; y<tilesy; y++){
        for(x=0; x<tilesx; x++){
            t[x][y].r.topLeft.x = pfleft + x*twidth;
            t[x][y].r.topLeft.y = PFTOP + y*thight;
            t[x][y].r.extent.x = twidth - 1;
            t[x][y].r.extent.y = thight - 1;
            
            if(d==tilesx*tilesy && newpf)
                d = 0;

            if(newpf)                   
                t[x][y].digit = d++;
            
            // because x+1 tile is not initialized at this time
            // when loading from a database
            if(t[x][y].digit != 0)		
                DrawTile(x, y, t[x][y].digit);

            if(t[x][y].digit == 0){
                holex = x;
                holey = y;
            }

        }
    }
}

//*****************************************************************
// DrawTile
// draws a tile with frame and lable or deletes one if d=0
//*****************************************************************

static void DrawTile(int x, int y, int d)
{
int fpx, fpy, slen;
char s[4];

	if(d>0){
	    DrawTileFrame(x, y);
	    StrIToA(s, d);
	    slen = StrLen(s);
	    fpx =  pfleft + x*twidth + (twidth - FntCharsWidth(s, slen))/2;
	    fpy = PFTOP + y*thight + (thight - FntCharHeight())/2;
	    WinDrawChars(s, slen, fpx,  fpy);
    }
	else{
        WinEraseRectangleFrame(roundFrame, &t[x][y].r);
		WinEraseRectangle(&t[x][y].r, 4);
		if(x-1 >= 0) DrawTileFrame(x-1, y);
		if(y-1 >= 0) DrawTileFrame(x, y-1);
		if(x+1 < tilesx) DrawTileFrame(x+1, y);
		if(y+1 < tilesy) DrawTileFrame(x, y+1);
	}
}

//*****************************************************************
// DrawTileFrame
// at x, y
//*****************************************************************

static void DrawTileFrame(int x, int y)
{
    WinDrawRectangleFrame(roundFrame, &t[x][y].r);
}

//*****************************************************************
// TilesAskSize
// opens a requester for a new tile x tile size of the playfield
// and initiates the new playfield
//*****************************************************************

#define LOWX    120

static void TilesAskSize()
{
FormPtr  form;
FieldPtr fptr_size;
char zwi[3];
Word obin_size;
VoidHand memhandle;
VoidPtr memptr;
CharPtr xpoint;

    memhandle = MemHandleNew(6);    // xxXyy/0 = 6 byte
    memptr = MemHandleLock(memhandle);
    StrIToA(memptr, tilesx);
    StrCat(memptr, "x");
    StrCat(memptr, StrIToA(zwi , tilesy));

    form = FrmInitForm(formID_size);
    obin_size = FrmGetObjectIndex(form, fieldID_size);
    fptr_size = FrmGetObjectPtr(form, obin_size);
    FldSetTextHandle(fptr_size, (Handle)memhandle);
    FrmSetFocus(form, obin_size);
    FrmDoDialog(form);

    if(FldGetTextLength(fptr_size) && (xpoint = StrChr(memptr, LOWX))){
        tilesx = StrAToI(memptr);
        tilesy = StrAToI(xpoint + 1);
    }

    if(tilesx < 2)  tilesx = 2;
    if(tilesx > MAXS) tilesx = MAXS;
    if(tilesy < 2)  tilesy = 2;
    if(tilesy > MAXS) tilesy = MAXS;

    MemHandleUnlock(memhandle);
    MemHandleFree(memhandle);

    newpf = true;
    GenPlayField();
}


//*****************************************************************
// TilesMix
// randomly mixes the tiles
//*****************************************************************

static void TilesMix()
{
int rx, ry, i;
SWord pScreenX, pScreenY;
Boolean pPenDown;

	do{
		EvtGetPen(&pScreenX, &pScreenY, &pPenDown);

        while((rx=SysRandom(0)&0x0f) >= tilesx)
            ;
        ry = holey;
        DetMoveTiles(rx, ry);
        
        while((ry=SysRandom(0)&0x0f) >= tilesy)
            ;
        rx = holex;
        DetMoveTiles(rx, ry);

   }while(!pPenDown); 
}

//*****************************************************************
// DetTile
// determines which tile is hit by pen and passes on this
// information to DetMoveTiles
//*****************************************************************

static void DetTile(int xp, int yp)
{
int x, y, x1, y1, x2, y2;

    for(x=0; x<tilesx; x++){
        x1 = t[x][0].r.topLeft.x;
        x2 = x1 + t[x][0].r.extent.x;
        if(x1<xp && x2>xp){
            for(y=0; y<tilesy; y++){
                y1 = t[0][y].r.topLeft.y;
                y2 = y1 + t[0][y].r.extent.y;
                if(y1<yp && y2>yp){
                    DetMoveTiles(x, y);
                    break;
                }
            }
            break;
        }
    }
}

//*****************************************************************
// DetMoveTiles
// if the selected tile is movable, all tiles from the selected
// one up to hole are moved
//*****************************************************************

static void DetMoveTiles(int px, int py)
{
    if(holex==px ^ holey==py){
    	do{
    		MoveTile(px, py);
    	}while(px!=holex || py!=holey);
    }
}

//*****************************************************************
// MoveTile
// delete the tile next to hole, mark new hole there and
// draw a tile at old hole position
//*****************************************************************

static void MoveTile(int px, int py)
{
int x, y, dx, dy;

	dx = holex - px;
	dy = holey - py;
    x = px + dx - sign(dx);
    y = py + dy - sign(dy);

	DrawTile(x, y, 0);
	DrawTile(holex, holey, t[x][y].digit);
	t[holex][holey].digit = t[x][y].digit;
	t[x][y].digit = 0;
	holex = x;
	holey = y;
}

//*****************************************************************
// sign
// returns -1/0/1 for negative/zero/positive values
//*****************************************************************

static int sign(int z)
{
    return(z==0 ? 0 : (z>0 ? 1 : -1));
}